from abc import ABC, abstractmethod
from qiskit import QuantumCircuit, QuantumRegister
import qiskit
import numpy as np

from .utility import Gate


# ------------> Encode Base Class <------------
class Encode(ABC):
    """
    Base class for encoding. Derived classes must overwrite the abstract methods.
    """

    def __init__(self):
        pass

    @abstractmethod
    def circuit(self, x):
        pass

    @abstractmethod
    def n_qubits(self, x):
        pass

    def all_circuits(self, D):
        """
        Enocde a dataset. Returns a list of circuits. One for each data exmaple (row).

        Input:
            - D (np.npdarray): A dataset. Rows represent data examples

        Returns:
            - (np.ndrray of qiskit.QuantumCircuit): An array of Quantum Circuits.
            The circuit at index i encodes data example at row i in D.
        """

        circuit_list = []

        for x in D:
            circuit_list.append(self.circuit(x))

        return circuit_list

    @staticmethod
    def print_encoders():
        """
        Prints the available derived classes.
        """

        print("AngleEncoding \n",
              "BasisEncoding \n",
              "BinaryPhaseEncoding \n",
              "DenseAngleEncoding \n")


# ------------> Angle Encoding <------------
class AngleEncoding(Encode):
    """
    Angle Encoding class. Assumes data is feature-normazized.
    """

    def __init__(self, gate=Gate.RY, scaling=np.pi/2):
        """
        Attributes:
            - gate=Gate.RY (Qiskit Gate): The rotation gate
            - scaling=np.pi/2 (float): Number by which to scale normalized input data.
            The defualt scaling is pi/2 which does not induce a relative phase
            difference.
        """

        self.gate = gate
        self.scaling = scaling

    def n_qubits(self, x):
        """
        Input:
            - x (np.ndarray): The input data to encode

        Returns:
            - Number of qubits needed to encode x
        """

        return len(x)

    def circuit(self, x):
        """
        Input:
            - x (np.ndarray): The input data to encode

        Returns:
            - (qiskit.QuantumCircuit): The circuit that encodes x

        Assumes data is feature-normalized. Assumes every element in x should be in [0, 1].
        """

        n_qubits = self.n_qubits(x)

        Sx = QuantumCircuit(n_qubits)

        for i in range(n_qubits):
            Sx.append(self.gate(2 * self.scaling * x[i]), [i])

        return Sx

    def state_vector(self, x):
        """
        Input:
            - x (np.ndarray): The input data to encode

        Returns:
            - np.array: The state vector representation of x after angle encoding
        """

        from functools import reduce

        gate_fn = lambda x: self.gate(2 * self.scaling * x).to_matrix()[:, 0]
        qubit_states = list(map(gate_fn, x[::-1]))

        """
        # Leaving this here because it might be more readable. Does the same
        # as the above code.
        qubit_states = []
        for x_i in x:
            qubit_state =  self.gate(2 * self.scaling * x_i).to_matrix()[:, 0]
            qubit_states.append(qubit_state)
        """
        return reduce(lambda a, b: np.kron(a, b), qubit_states)


# ------------> Dense Angle Encoding <------------
class DenseAngleEncoding(Encode):
    """
    Dense Angle Encoding Class.

    Assumptions:
    - Assumes data is feature-normalized
    - Assumes the number of featers in x is divisible by two. If this is not the
    case for your data, consider appending zeros.
    """

    def __init__(self, rotation_gate=Gate.RY, scaling=np.pi/2):
        """
        Attributes:
            gate=Gate.RY (Qiskit Gate): The rotation gate
            scaling=np.pi/2 (float): Number by which to scale normalized input data.
            The defualt scaling is pi/2 which does not induce a relative phase.
        """

        self.rotation_gate = rotation_gate
        self.scaling = scaling

    def n_qubits(self, x):
        """
        Assumes the number of features in x is divisible by two

        Input:
            - x (np.ndarray): The input data to encode

        Returns:
            - Number of qubits needed to encode x
        """

        assert_string = "DenseAngleEncoding assumes the number of features in x is divisible by two"
        assert (len(x) / 2) % 1 == 0, assert_string

        return len(x) // 2

    def circuit(self, x):
        """
        Input:
            - x (np.ndarray): The input data to encode

        Returns:
            - (qiskit.QuantumCircuit): The circuit that encodes x

        Assumes data is feature-normalized.
        Assumes every element in x should be in [0, 1].
        """

        n_qubits = self.n_qubits(x)

        Sx = QuantumCircuit(n_qubits)

        for i in range(n_qubits):
            rotation_idx = 2*i
            phase_idx = 2*i + 1

            phase_factor = 2*np.pi * x[phase_idx]

            Sx.append(self.rotation_gate(2 * self.scaling * x[rotation_idx]), [i])
            Sx.u1(phase_factor, [i])

        return Sx

    def state_vector(self, x):
        """
        Input:
            - x (np.ndarray): The input data to encode

        Returns:
            - np.array: The state vector representation of x after dense angle encoding
        """

        from functools import reduce

        qubit_states = []
        for i in range(len(x)//2):
            x_i = x[2*i] # qubit that defines the rotation
            x_j = x[2*i + 1] # qubit that defines the phase
            rotation = self.rotation_gate(2 * self.scaling * x_i).to_matrix()[:, 0]
            phase_and_rotation = [rotation[0], np.exp(2 * np.pi * 1j *x_j) * rotation[1]]
            qubit_states.append(phase_and_rotation)

        qubit_states = qubit_states[::-1] # match qiskit qubit ordering

        return reduce(lambda a, b: np.kron(a, b), qubit_states)


# ------------> Binary Phase Encoding <------------
class BinaryPhaseEncoding(Encode):
    """
    Binary Phase Encoding Class.

    Assumptions:
        - Assumes binary data with each feature in {-1, 1}
        - Assumes the number of featers in x is a power of two
        case for your data, consider appending zeros.
    """

    def __init__(self, method="SF", ancilla=True):
        """
        Attributes:
            - method="SF": For sign-flip algorithm.
            - ancilla=True (Boolean): Add ancillary qubit to circuit

        Hypergraph state generation as a method in a future release
        """

        self.method = method
        self.ancilla = ancilla

    def n_qubits(self, x, plus_ancilla=False):
        """
        Input:
            - x (np.ndarray): The input data to encode
            - plus_ancilla=False (Boolean): Count ancilla qubits if self.ancilla=True

        Returns:
            - Number of qubits needed to encode x. Includes ancilla qubit if
            self.ancilla is true and plus_ancilla=True

        Assumes the number of features in x is a power of 2.
        """

        assert_string = "BinaryPhaseEncoding assumes the number of features in x is a power of 2"
        assert np.log2(len(x)) % 1 == 0, assert_string

        n_qubits = int(np.log2(len(x)))
        if self.ancilla == True and plus_ancilla == True:
            n_qubits += 1

        return n_qubits


    def circuit(self, x, ancilla=None):
        """
        Input:
            - x (np.ndarray): The input data to encode.
            - ancilla=None (Boolean): If None, uses self.ancilla.
            Adds an ancillary qubit to the circuit, if true.

        Returns:
            - (qiskit.QuantumCircuit): The circuit that encodes x

        Assumptions:
            - Assumes binary data with each feature in {-1, 1}
            - Assumes the number of featers in x is a power of two
            case for your data, consider appending zeros.
        """

        if ancilla is None:
            ancilla = self.ancilla

        if self.method == "SF":
            return self.SF(x, ancilla)
        elif self.method == "HSGS":
            raise NotImplementedError
        else:
            raise ValueError


    def SF(self, x, ancilla):
        """
        Sign-Flip algorithm for data encoding

        Input:
            - x (np.ndarray): The input data to encode
            - ancilla=True (Boolean): Adds an ancillary qubit to the circuit, if true

        Returns:
            - (qiskit.QuantumCircuit): The circuit that encodes x using the SF algorithm


        Assumptions:
            - Assumes binary data with each feature in {-1, 1}
            - Assumes the number of featers in x is a power of two
            case for your data, consider appending zeros.
        """

        x = np.array(x)
        d = len(x) # dimensionality of data vector
        n_qubits_ancilla = self.n_qubits(x, ancilla) # qubits plus ancilla qubit if self.ancilla=True

        if ancilla == True:
            n_qubits = n_qubits_ancilla - 1 # non ancilla qubits
        else:
            n_qubits = n_qubits_ancilla

        Sx = QuantumCircuit(n_qubits_ancilla)
        qubit_idx_list = list(range(n_qubits))

        # Generate computational basis vectors in Dirac notation
        basis_labels = [("{:0%db}"%n_qubits).format(k) for k in range(d)]

        # Create multi-controlled Z gate, or single Z gate if N = 1qubit.
        Z = qiskit.circuit.library.standard_gates.ZGate()
        if n_qubits == 1:
            z_op = Z
        else:
            z_op = Z.control(n_qubits-1)

        # Full layer of H
        Sx.h(qubit_idx_list)

        # Find all components with a -1 factor in i (and thus our target state vector)
        indices = np.where(x == -1)[0]
        if indices.size > 0:
            for idx in indices:
                # Need to switch qubits in the 0 state so CZ will take effect
                for i, b in enumerate(basis_labels[idx]):
                    if b == '0':
                        Sx.x((n_qubits-1)-i) # (N-1)-i is to match the qubit ordering Qiskit uses (reversed)

                Sx.append(z_op, qubit_idx_list)

                # And switch the flipped qubits back
                for i, b in enumerate(basis_labels[idx]):
                    if b == '0':
                        Sx.x((n_qubits-1)-i)

        return Sx


class BasisEncoding(Encode):
    """
    Basis Encoding Class.

    Encodes binary vectors (can be thought of as bit strings) into the state
    with the corresponding label in Dirac notation.
    """

    def __init__(self):
        pass

    def n_qubits(self, x):
        """
        Input:
            - x (np.ndarray): The input data to encode

        Returns:
            - Number of qubits needed to encode x.
        """

        return len(x)

    def circuit(self, x):
        """
        Input:
            - x (np.ndarray): The input data to encode

        Returns:
            - (qiskit.QuantumCircuit): The circuit that encodes x

        Assumptions:
            - Assumes binary data with each feature in {0, 1}
        """

        if isinstance(x, list):
            x = np.array(x)

        assert (np.count_nonzero(x == 0) + np.count_nonzero(x == 1)) == len(x), "All features must be {0, 1}"


        x = np.array(x)
        x_reversed = x[::-1] # match Qiskit qubit ordering

        n_qubits = self.n_qubits(x)
        Sx = QuantumCircuit(n_qubits)

        one_indices = np.where(x_reversed == 1)[0]
        for i in one_indices:
            Sx.x(i)

        return Sx



import numpy as np
import qiskit
from abc import ABC, abstractmethod
from .observable import Observable

class Measurement(ABC):
    """
    Abstract class with one abstract method, output(). Derived classes must overwrite this method. This class serves
    as the output layer. The output function takes the results of running a circuit transforms the results into the
    final output.

    To print the default Derived classes you can call Measurement.print_derived_classes().
    """

    def __init__(self, qubits, rotate=False):
        """
        Attributes:
            - self.qubits (list or np.array): The qubits to be measured
            - self.rotate=False (boolean): True if the measurement is performed
            with respect to a basis other than the computational basis
        """
        self.qubits = qubits
        self.rotate = rotate

    @abstractmethod
    def output(counts):
        """
        Overwrite this method in dervided classes.

        Input:
            - counts (dict): The result of running the circuit

        Returns:
            - (np.ndarray): Model output after measurement tranformations
        """
        pass

    @staticmethod
    def born_rule(amplitude):
        """
        Input:
            - amplitude (float): complex amplitude
        Returns:
            - |amplitude|^2 (float): The probability of being measured in the state
            corresponding to the amplitude
        """

        return abs(amplitude)**2

    @staticmethod
    def statevector_to_probability(psi):
        """
        Input:
            - psi (np.ndarray):
        Returns:
            - (np.ndarray) of probabilities
        """

        probs = np.array([Measurement.born_rule(amp) for amp in psi])

        return probs

    @staticmethod
    def statevector_to_probability_dict(psi):
        """
        Input:
            - psi (np.ndarray):
        Returns:
            - Dictionary of measurment outcomes as keys and their corresponding probabilities as values
        """

        n_qubits = int(np.log2(len(psi)))
        bit_strings =  Measurement.get_bit_strings(n_qubits)

        prob_dict = {key: Measurement.born_rule(value) for key, value in zip(bit_strings, psi)}

        return prob_dict


    @staticmethod
    def counts_to_probability(counts):
        """
        Inputs:
            - counts (dict): Of the form {"measurement result": count}
        Returns:
            - (dict): Of the form {"measurement result": probability}
        """

        n_shots = sum(counts.values())

        prob_dict = {key: value/n_shots for key, value in counts.items()}

        return prob_dict


    @staticmethod
    def counts_for_qubit_subset(counts, qubits, bit_strings, qubit_dict):
        """
        counts_for_qubits() calls this function if the number of measured qubits is greater
        than the length of the qubits argument. I.e. more qubits were measured than we are
        interested in getting the counts for.

        Inputs:
            - counts (dict): Dictionary with counts as values and corresponding measurement
            results as keys.
            - qubits Union(list or int): qubit index or list of qubit indices

        Returns:
            - A dictionary with the keys being all possible bit strings corresponding to the sorted
            qubit indices and values being the counts for the measument outcomes for those qubit results.

        Example:
            - qubits=[0, 4] would return a dictionary of {'00': a, '01': b, '10': c, '11': d}
            Where the bit string is the state of qubit 4 (left) and qubit 0 (right), and the values
            are the counts.
        """
        sorted_qubits = list(reversed(sorted(qubits)))
        for key, value in counts.items():
            # Look at only the revelant qubits
            reverse_position = len(key) - 1
            qubit_key = "".join([key[reverse_position - sq] for sq in sorted_qubits])

            qubit_dict[qubit_key] += value

        return qubit_dict

    @staticmethod
    def counts_for_qubits(counts, qubits):
        """
        Inputs:
            - counts (dict): Dictionary with counts as values and corresponding measurement
            results as keys.
            - qubits Union(list or int): qubit index or list of qubit indices.

        Returns:
            - A dictionary with the keys being all possible bit strings corresponding to the sorted
            qubit indices and values being the counts for the measument outcomes for those qubit results.

        Example:
            - qubits=[0, 4] would return a dictionary of {'00': a, '01': b, '10': c, '11': d}
            Where the bit string is the state of qubit 4 (left) and qubit 0 (right), and the values
            are the counts.
        """
        if isinstance(qubits, int):
            qubits = [qubits]

        bit_strings = Measurement.get_bit_strings(len(qubits))

        # Create qubit dictionary
        qubit_dict = {}
        for s in bit_strings:
            qubit_dict[s] = 0

        n_measured_qubits = len(list(counts.keys())[0])
        if n_measured_qubits > len(qubits):
            return Measurement.counts_for_qubit_subset(counts, qubits,
                                                       bit_strings, qubit_dict)
        for key, value in counts.items():
            qubit_dict[key] += value

        return qubit_dict


    @staticmethod
    def rotate_basis(circuit, qubit, observable):
        """
        Note: modifies the circuit in-place

        Input:
        - circuit (qiskit.QuantumCircuit)
        - qubit (int): qubit index
        - observable (Observable)

        Description:
            Appends the necessary gates to the circuit so that eigenbasis of the observable is rotated
        to the computational basis.
        """

        if isinstance(qubit, int):
            qubit = [qubit]

        if observable.name == "X":
            circuit.h(qubit)

        elif observable.name == "Y":
            # Alternative is the commented version
            #circuit.z(qubit)
            #circuit.s(qubit)
            #circuit.h(qubit)
            circuit.sdg(qubit)
            circuit.h(qubit)

        elif observable.name == "Z":
            pass

        elif observable.name == "H":
            circuit.ry(np.pi/4, qubit)

        else:
            circuit.unitary(observable.eigenvectors.T.conj(), qubit)


    @staticmethod
    def get_bit_strings(n_qubits):
        """
        Input:
            - n_qubits (int):

        Returns:
            - A list of all combinations of bit strings of length n_qubits
        """
        # Generate all bit strings
        bit_strings = []

        for k in range(2**n_qubits):
            bit_strings.append(("{:0%db}" % n_qubits).format(k))

        return bit_strings

    @staticmethod
    def add_measurements(circuit, qubits=None, clbits=None, add_classical_register=True):
        """
        Adds measurements operations to circuit in-place. If no argument is passed for clbits
        and no classical register was added, the measurement result for the qubit at position
        i in the qubits list will be stored in the classical bit i. If no arguement is passed
        for clbits and a classical register was added the measurent result of qubit at list
        position i will be found in classical bit
        (i + number of classical bits originally in the circuit). This prevents overwritting
        existing classical bits.

        Inputs:
            - circuit (qiskit.QuantumCircuit)
            - qubits=None (list): Qubit indices to measure. If none it is the ordered range of all qubits.
            - clbits=None (list): Classical bit indices to hold the labeled outcome of qubit measurements.
            - add_classical_register=True (boolean): If true, adds a classical register of size
            len(qubits) to circuit
        """
        if qubits == None:
            qubits = list(range(circuit.num_qubits))

        n_measurements = len(qubits)
        n_existing_clbits = circuit.num_clbits
        if add_classical_register == True:
            cr = qiskit.ClassicalRegister(n_measurements)
            circuit.add_register(cr)

        if clbits == None:
            clbits = list(range(n_existing_clbits, n_existing_clbits + n_measurements))
            circuit.measure(qubits, clbits)
        else:
            circuit.measure(qubits, clbits)

    @staticmethod
    def print_derived_classes():
        print("Probability()")
        print("Expectation()")
        print("ProbabilityThreshold()")


class Probability(Measurement):
    """
    A class to transform circuit outputs into qubit probabilites.
    """

    def __init__(self, qubits, p_zero=True, observable=None):
        """
        Attributes:
            - qubits Union(int, list, np.ndarray): qubit index or list of qubit indices
            - observable_basis (Observable): The observable corresponding the basis to measure in
            - zero=True (Boolean): If True, output returns probabilties of qubit being measured in the |0> state.
            If false, output returns probabilties of qubit being measured in the |1> state.
        """

        if isinstance(qubits, int):
            self.qubits = [qubits]
        else:
            self.qubits = qubits
        self.p_zero = p_zero
        self.observable = observable
        if observable is not None:
            requires_rotation = True
        else:
            requires_rotation = False

        super().__init__(qubits=self.qubits, rotate=requires_rotation)


    def rotate_basis(self, circuit, qubit=None, observable=None):
        """
        Call rotate basis in super class, Measurement. Modifies circuit in-place.
        Uses attributes of self for optional arguments which are not provided.
        Input:
            - circuit (qiskit.QuantumCircuit)
            - qubit=None (int or list)
            - observable (QNN-Gen Observable)
        """

        if qubit is None:
            qubit = self.qubits

        if observable is None:
            observable = self.observable

        super().rotate_basis(circuit, qubit, observable)


    def output(self, counts):
        """
        Input:
            - counts (dict):

        Returns:
            - A list (np.ndarray) of probabilities for each qubit.

        probability_list[0] is the probability of the the qubit whose index is given by
        self.qubits[0] being in the |1> state (if self.one = True)
        """

        probability_list = []

        for qubit in self.qubits:
            qubit_counts = self.counts_for_qubits(counts, qubit)
            prob_dict = self.counts_to_probability(qubit_counts)

            # Note: below might not be general enough for multi-qubit measurements
            if self.p_zero == True:
                probability_list.append(prob_dict['0'])
            else:
                probability_list.append(prob_dict['1'])

        return np.array(probability_list)


class ProbabilityThreshold(Measurement):
    """
    The output function uses a threshold on probabilities of being in the zero state (if zero=True)
    to return a binary classification result.

    E.g. If threshold=0.5 and the probability of the qubit being in the |0> state is >= 0.5 then
    the output function will return a label of [0] for this qubit. If the probability of the qubit
    being in the |0> state is < 0.5 then a label of [1] is returned by output() by default.
    """

    def __init__(self, qubits, p_zero=True, threshold=0.5, labels=None, observable=None):
        """
        Attributes:
        - qubits (int or list): qubit index  or list of qubit indices

        - observable (QNN-Gen Observable): The observable corresponding the basis to measure in

        - p_zero=True (Boolean): If True, output returns probabilties of qubit being measured in the |0> state.
        If false, output returns probabilties of qubit being measured in the |1> state.

        - threshold=0.5 (float): For binary classification. Should be between 0 and 1.

        - labels=None (np.ndarray): The lables to return from output. A 2 element list. labels[0] is the label
        corresponding to a probability that execedes the threshold, labels[1] corresponds to the probability
        being
        """

        if isinstance(qubits, int):
            self.qubits = [qubits]
        else:
            self.qubits = qubits

        self.p_zero = p_zero
        self.threshold = threshold

        if labels is None:
            if p_zero == True:
                self.labels = np.array([0, 1])
            else:
                self.labels = np.array([1, 0])
        else:
            self.labels = labels

        self.observable = observable
        if observable is not None:
            requires_rotation = True
        else:
            requires_rotation = False

        super().__init__(qubits=self.qubits, rotate=requires_rotation)


    def rotate_basis(self, circuit, qubit=None, observable=None):
        """
        Call rotate basis in super class, Measurement. Modifies circuit in-place.
        Uses attributes of self for optional arguments which are not provided.
        Input:
            - circuit (qiskit.QuantumCircuit)
            - qubit=None (int or list)
            - observable (QNN-Gen Observable)
        """
        if qubit is None:
            qubit = self.qubits

        if observable is None:
            observable = self.observable

        super().rotate_basis(circuit, qubit, observable)


    def output(self, counts):
        """
        Input:
            - counts (dict):
        Returns:
            - (np.ndarray): The element of self.label corresponding to the outcome after thresholding
        """
        prob = Probability(self.qubits, self.p_zero, self.observable)
        prob_vec = prob.output(counts)
        threshold_vec = prob_vec >= self.threshold # a boolean vector
        labels = np.where(threshold_vec, self.labels[0], self.labels[1]) # labels[0] where true

        return labels


class Expectation(Measurement):
    """
    A class to assist in computing expectation values. Currently for one qubit expectation.
    """

    def __init__(self, qubits, observable=Observable.Z()):
        """
        - qubit (int): The index of the qubit
        - observable=Observable.Z() (matrix, tuple, or QNN-Gen Observable object): The observable
        to meausre with respect to
            - If matrix it should be a list or numpy.ndarray
            - If tuple, it should be of the form ("name": matrix)
        """

        if isinstance(qubits, int):
            self.qubits = [qubits]
        else:
            self.qubits = qubits # just single-qubit for now

        if isinstance(observable, (list, np.ndarray)):
            self.observable = Observable(observable)

        elif isinstance(observable, tuple):
            self.observable = Observable(matrix=observable[1], name=observable[0])

        elif isinstance(observable, Observable):
            self.observable = observable

        else:
            raise ValueError("Observable argument is not a correct type. Check the __init__ docstring.")

        if self.observable.name == "Z":
            requires_rotation = False
        else:
            requires_rotation = True

        super().__init__(qubits=self.qubits, rotate=requires_rotation)

    def rotate_basis(self, circuit, qubit=None, observable=None):
        """
        Overwritting rotate_basis in Measurement
        """

        if qubit is None:
            qubit = self.qubits

        if observable is None:
            observable = self.observable

        super().rotate_basis(circuit, qubit, observable)

    def output(self, counts):
        """
        Input:
            - counts (dict):
        Returns:
            - (np.ndarray):
        """
        # Convert to probability
        qubit_counts = self.counts_for_qubits(counts, self.qubits)
        prob_dict = self.counts_to_probability(qubit_counts)

        # Match measurement results to corresponding eigenvalues
        prob_tuples = tuple(prob_dict.items())
        sorted_prob_tuple = sorted(prob_tuples, key=lambda k: int(k[0], 2)) # sort by bits
        # sorting is to make list positions align with self.observabale.eigenvalues

        probs = np.array([prob for (key, prob) in sorted_prob_tuple])

        return np.array([np.dot(self.observable.eigenvalues, probs)])



"""
Will need to import Measurement for default_measurements
"""

from abc import ABC, abstractmethod
from qiskit import QuantumCircuit, QuantumRegister
import qiskit
import numpy as np
from .utility import Gate

class Model(ABC):
    """
    Class to define a QML model, whether it be a particular ansatz or a perceptron model.

    Abstract class. Derived classes overwrite circuit() and default_measuremnt(). Derived classes should
    also have a self.params attribute.
    """

    def __init__(self):
        """
        Attributes:
            - n_qubits (int): Number of encoded qubits in the encoding circuit to be joined with
            the model circuit. In other words, number of qubits not counting ancillary qubits.
            - params (np.ndarray): The model's learnable parameters
        """

        self.n_qubits = None
        self.parameters = None

    @abstractmethod
    def circuit():
        """
        Abstract method. Overwrite to return the circuit.

        Returns:
            - (qiskit.QuantumCircuit): The circuit defined by the model
        """
        pass


    def default_measurement():
        """
        Often, the model implies what measurements are sensible.

        Returns:
            - (Measurement): The default measurement for the model
        """

        raise NotImplementedError

    @staticmethod
    def print_models():
        """
        Prints the available derived classes.
        """

        print("TensorTreeNetwork" + "\n",
              "BinaryPerceptron")


class TreeTensorNetwork(Model):
    """
    Class to implement Tensor Tree Networks.

    Assumptions:
        - Number of features is a power of two.
    """

    def __init__(self, n_qubits=None, angles=None, rotation_gate=Gate.RY, entangling_gate=Gate.CX):
        """
        Inputs:
            - n_qubits=None (int): Number of encoded qubits
            - anlges=None (list): Parameters for rotation_gate. Sets self.parameters
            - rotation_gate=Gate.RY (qiskit.circuit.library.standard_gates): 1-qubit rotation gate
            - entangling_gate=Gate.CX (qiskit.circuit.library.standard_gates): 2-qubit entangling gate
        """

        self.n_layers = None
        self.measurement_qubit = None
        self.parameters = angles
        self.n_qubits = n_qubits # may update attributes above
        self.rotation_gate = rotation_gate
        self.entangling_gate = entangling_gate

    @property
    def n_qubits(self):
        return self._n_qubits

    @n_qubits.setter
    def n_qubits(self, n_qubits):
        """
        Sets self.n_qubits. Assumes n_qubits is a power of two.

        Also may update self.n_layers, self.measurement_qubit, and self.angles
        """

        if n_qubits == None:
            self._n_qubits = None
            return

        if np.log2(n_qubits) % 1 != 0:
            import warnings
            warnings.warn("TTN assumes n_qubits is a power of two.")

        self._n_qubits = n_qubits

        self.n_layers = int(np.log2(n_qubits))
        self.measurement_qubit = n_qubits // 2

        if self.parameters is None:
            self.parameters = self.get_rand_angles(n_qubits)
        else:
            if int(2**(self.n_layers + 1)) != len(self.parameters):
                import warnings
                warnings.warn("""Number of quibts no longer match the number of parameters.
                              To update parameters use set_params(new_angles) or
                              self.get_rand_angles(n_qubits).""")


    def get_rand_angles(self, n_qubits):
        """
        Generates random numbers from [0, pi)
        """

        n_parameters = 2**(self.n_layers + 1) - 2

        rand_angles = np.random.rand(n_parameters) * np.pi

        return rand_angles


    def circuit(self):
        """
        Returns:
            - qiskit.QuantumCircuit
        """
        n_qubits = self.n_qubits
        angles = self.parameters
        qc = QuantumCircuit(n_qubits)

        def layer(angle_idx):
            qubit_pairs = [(active_qubits[2*i], active_qubits[2*i+1]) for i in range(int(len(active_qubits)/2))]

            next_active = []

            for i, q in enumerate(qubit_pairs):
                qc.append(self.rotation_gate(angles[angle_idx]), [q[0]])
                qc.append(self.rotation_gate(angles[angle_idx + 1]), [q[1]])

                if q[0] < self.measurement_qubit:
                    qc.append(self.entangling_gate, [q[0], q[1]])
                    next_active.append(q[1])
                else:
                    qc.append(self.entangling_gate, [q[1], q[0]])
                    next_active.append(q[0])

                angle_idx += 2

            return next_active, angle_idx

        active_qubits = np.arange(n_qubits) # qubit indices
        n_active = len(active_qubits)

        angle_idx = 0
        while n_active >= 2:
            active_qubits, angle_idx = layer(angle_idx)
            n_active = len(active_qubits)

        return qc


    def default_measurement(self, observable=None, measurement_qubit=None):
        """
        Default measurement is the Z expectation of the measurement_qubit

        Inputs:
            - observable=None (Observable): If None, defaults to Observable.Z()
            - measurement_qubit (int): If None, defaults to self.measurement_quibt = n_qubits//2

        Returns:
            - Expectation() measurement object
        """
        from .observable import Observable
        from .measurement import Expectation

        if observable is None:
            observable = Observable.Z()

        if measurement_qubit is None:
            measurement_qubit = self.measurement_qubit

        expectation = Expectation(measurement_qubit, observable)

        return expectation


class BinaryPerceptron(Model):
    """
    Class to implement Binary Perceptron models.

    Assumptions:
        - Number of features is a power of two.
        - Features are binary elements in {-1, 1}
    """

    def __init__(self, n_qubits=None, weights=None):
        """
        Attribuites:
            - n_qubits (int)
            - weights (np.ndarray): A binary vector with elements in {-1, 1}
        """

        # Ensure weights is a np array if passed
        if weights is not None:
            self.parameters = np.array(weights)
        else:
            self.parameters = weights
        self.measurement_qubit = None
        self.n_qubits = n_qubits

    @property
    def n_qubits(self):
        return self._n_qubits

    @n_qubits.setter
    def n_qubits(self, n_qubits):
        """
        Input:
            n_qubits (int): number of encoded qubits

        Sets self.n_qubits. Assumes n_qubits is a power of two.
        Also may update self.n_layers, self.measurement_qubit, and self.angles
        """

        if n_qubits == None:
            self._n_qubits = None
            return

        #n_qubits -= 1 # Not counting ancilla
        assert np.log2(n_qubits) % 1 == 0, "Assumes n_qubits is a power of two."
        self._n_qubits = n_qubits

        if self.measurement_qubit is None:
            if n_qubits == 1:
                self.measurement_qubit = 0
            else:
                self.measurement_qubit = n_qubits

        if self.parameters is None:
            self.parameters = self.init_random_weights(self.n_qubits)
        else:
            if int(2**n_qubits) != len(self.parameters):
                import warnings

                warnings.warn("""Number of quibts no longer match the number of parameters.
                              To update parameters use set_params(new_weights) or
                              init_random_weights(n_qubits).""")

    def init_random_weights(self, n_qubits):
        return np.random.choice([-1, 1], 2**n_qubits)

    def circuit(self):
        """
        Returns the circuit for the model.
        """

        w = self.parameters

        # Not including ancilla
        N = self.n_qubits
        qubit_idx_list = list(range(N))

        # Generate computational basis vectors in Dirac notation
        basis_labels = [("{:0%db}"%N).format(k) for k in range(len(w))]

        # No need for ancilla if N = 1
        if N == 1:
            Sx = QuantumCircuit(N)
            qubit_ancilla_idx_list = qubit_idx_list

        else:
            Sx = QuantumCircuit(N + 1) # + 1 for ancilla

            qubit_ancilla_idx_list = qubit_idx_list + [N] # [N] is the index of the ancilla

        # Create multi-controlled Z gate, or single Z gate if N = 1qubit. Likewise for X
        Z = qiskit.circuit.library.standard_gates.ZGate()
        X = qiskit.circuit.library.standard_gates.XGate()
        if N == 1:
            z_op = Z
            x_op = X

        else:
            z_op = Z.control(N-1)
            x_op = X.control(N)

        # Find all components with a -1 factor in i (and thus our target state vector)
        indices = np.where(w == -1)[0]
        if indices.size > 0:
            for idx in indices:
                # Need to switch qubits in the 0 state so CZ will take effect
                for i, b in enumerate(basis_labels[idx]):
                    if b == '0':
                        Sx.x((N-1)-i) # (N-1)-i is to match the qubit ordering Qiskit uses (reversed)

                Sx.append(z_op, qubit_idx_list)

                # And switch the flipped qubits back
                for i, b in enumerate(basis_labels[idx]):
                    if b == '0':
                        Sx.x((N-1)-i)

        Sx.h(qubit_idx_list)
        Sx.x(qubit_idx_list)
        Sx.append(x_op, qubit_ancilla_idx_list)

        return Sx


    def default_measurement(self, observable=None, measurement_qubit=None):
        """
        A measurement on the ancilla qubit. If the probability of the ancilla being in the |1> is
        greater the than the threshold, output gives label 1. Otherwise, label 0.

        Default threshold is 0.5.

        Returns
            - ProbabilityThreshold Measurement object.
        """

        from .measurement import ProbabilityThreshold

        if measurement_qubit == None:
            qubits = self.measurement_qubit

        pt = ProbabilityThreshold(qubits=qubits,
                                 p_zero=False,
                                 threshold=0.5,
                                 labels=[1, 0],
                                 observable=observable)

        return pt


class EntangledQubit(Model):
    """
    Class to an engtangled qubit classifier.
    """

    def __init__(self,
        n_qubits=None,
        angles=None,
        n_layers=3,
        gate_set=[Gate.RXX, Gate.RZX]):
        """
        Inputs:
            - n_qubits=None (int): Number of encoded qubits
            - anlges=None (list): Parameters for rotation gate
            - gate_set=[Gate.RXX, Gate.RZX] (list(qiskit gates)): Parameterized two qubit
            gates teo use. self.circuit() loop through the list of gates and repeat until
            n_layers have been constructed.
        """

        self.n_layers = n_layers
        self.measurement_qubit = 0
        self.parameters = angles
        if angles is None:
            self.n_parameters = None
        else:
            self.n_parameters = len(angles)
        self.n_qubits = n_qubits # may update attributes above
        self.gate_list = gate_set

    @property
    def n_qubits(self):
        return self._n_qubits

    @n_qubits.setter
    def n_qubits(self, n_qubits):
        """
        Sets self.n_qubits. Assumes n_qubits is a power of two.

        Also may update self.n_layers, self.measurement_qubit, and self.angles
        """

        if n_qubits == None:
            self._n_qubits = None
            return

        assert np.log2(n_qubits) % 1 == 0, "Assumes n_qubits is a power of two."
        self._n_qubits = n_qubits

        if self.parameters is None:
            self.n_parameters = (n_qubits - 1) * self.n_layers
            self.parameters = self.get_rand_angles(self.n_parameters)
        else:
            if int(2**(self.n_layers + 1)) != len(self.parameters):
                import warnings
                warnings.warn("""Number of quibts no longer match the number of parameters.
                              To update parameters use set_params(new_angles) or
                              self.get_rand_angles(n_qubits).""")

    def get_rand_angles(self, n_parameters):
        """
        Generates random numbers from [0, 2*pi)
        """

        rand_angles = np.random.uniform(0, 2*np.pi, n_parameters)

        return rand_angles

    def circuit(self):
        """
        Returns:
            - qiskit.QuantumCircuit
        """

        qc = QuantumCircuit(self.n_qubits)

        for layer_idx in range(self.n_layers):
            # Loop through gate list, return to beginning after end of the list
            gate = self.gate_list[layer_idx % len(self.gate_list)]

            for gate_idx in range(1, self.n_qubits):
                gate_parameter = self.parameters[layer_idx*(self.n_qubits-1) + gate_idx - 1]
                qc.append(gate(gate_parameter), [self.measurement_qubit, gate_idx])

        return qc

    def default_measurement(self, observable=None, measurement_qubit=None):
        """
        Default measurement is the Z expectation of the measurement_qubit

        Inputs:
            - observable=None (qnn_gen.Observable): If None, defaults to Observable.Y()
            - measurement_qubit (int): If None, defaults to self.measurement_qubit,
            which itself defaults to qubit 0.

        Returns:
            - Expectation() measurement object
        """
        from .observable import Observable
        from .measurement import Expectation

        if observable is None:
            observable = Observable.Y()

        if measurement_qubit is None:
            measurement_qubit = self.measurement_qubit

        expectation = Expectation(measurement_qubit, observable)

        return expectation



import qiskit.circuit.library.standard_gates
from qiskit import Aer
from qiskit import execute

def get_state_vector(circuit):
    """
    Input:
        - circuit (qiskit.QuantumCircuit)
    Returns:
        - state vector
    """
    backend = Aer.get_backend("statevector_simulator")
    psi = execute(circuit, backend).result().get_statevector()

    return psi

def get_counts(circuit, measure_all=True):
    """
    Input:
        - circuit (qiskit.QuantumCircuit)
        - measure_all=True (boolean): If true, adds measurement operations and corresponding
        classical registers for all qubits
    Returns:
        - dictionary of counts
    """
    if measure_all == True:
        circuit.measure_all()

    backend = Aer.get_backend("qasm_simulator")
    counts = execute(circuit, backend).result().get_counts(circuit)

    return counts

class Gate:
    """
    A class for easy access to Qiskit gates.
    """

    def __init__(self, gate_string):
        self.gate_string = gate_string.lower()
        self.gate = self.gate_parser()


    def gate_parser(self):
        if self.gate_string == "x":
            return Gate.X
        elif self.gate_string == "y":
            return Gate.Y
        elif self.gate_string == "z":
            return Gate.Z
        elif self.gate_string == "h":
            return Gate.H
        elif self.gate_string == "s":
            return Gate.S
        elif self.gate_string == "sdg":
            return Gate.Sdg
        elif self.gate_string == "t":
            return Gate.T
        elif self.gate_string == "tdg":
            return Gate.Tdg
        elif self.gate_string == "rx":
            return Gate.RX
        elif self.gate_string == "ry":
            return Gate.RY
        elif self.gate_string == "rz":
            return Gate.RZ
        elif self.gate_string == "u1":
            return Gate.U1
        elif self.gate_string == "u2":
            return Gate.U2
        elif self.gate_string == "u3":
            return Gate.U3
        elif self.gate_string == "cx":
            return Gate.CX
        elif self.gate_string == "cy":
            return Gate.CY
        elif self.gate_string == "cz":
            return Gate.CZ
        elif self.gate_string == "swap":
            return Gate.SWAP
        else:
            return "Invalid gate string"


    @staticmethod
    def print_gates():
        print("X\n",
              "Y\n",
              "Z\n",
              "H\n",
              "S\n",
              "Sdg\n",
              "T\n",
              "Tdg\n",
              "RX\n",
              "RY\n",
              "RZ\n",
              "U1\n",
              "U2\n",
              "U3\n",
              "CX\n",
              "CY\n",
              "CZ\n",
              "SWAP")

    # --------------------> one qubit gates <-----------------
    # Pauli
    X = qiskit.circuit.library.standard_gates.XGate()
    Y = qiskit.circuit.library.standard_gates.YGate()
    Z = qiskit.circuit.library.standard_gates.ZGate()

    # Hadamard
    H = qiskit.circuit.library.standard_gates.HGate()

    # Phase
    S = qiskit.circuit.library.standard_gates.SGate() # ~=Z**0.5
    Sdg = qiskit.circuit.library.standard_gates.SdgGate()

    T = qiskit.circuit.library.standard_gates.TGate() # ~=Z**0.25
    Tdg = qiskit.circuit.library.standard_gates.TdgGate()

    # Puali Rotations
    RX = lambda theta : qiskit.circuit.library.standard_gates.RXGate(theta)
    RY = lambda theta : qiskit.circuit.library.standard_gates.RYGate(theta)
    RZ = lambda theta : qiskit.circuit.library.standard_gates.RZGate(theta)

    # "U" gates
    U1 = lambda theta : qiskit.circuit.library.standard_gates.U1Gate(theta)
    U2 = lambda phi, lam : qiskit.circuit.library.standard_gates.U2Gate(phi, lam)
    U3 = lambda theta, phi, lam : qiskit.circuit.library.standard_gates.U2Gate(theta, phi, lam)

    # --------------------> two qubit gates <-----------------
    CX = qiskit.circuit.library.standard_gates.CXGate()
    CY = qiskit.circuit.library.standard_gates.CYGate()
    CZ = qiskit.circuit.library.standard_gates.CZGate()

    RXX = lambda theta : qiskit.circuit.library.standard_gates.RXXGate(theta)
    RZX = lambda theta : qiskit.circuit.library.standard_gates.RZXGate(theta)

    swap = qiskit.circuit.library.standard_gates.SwapGate()



from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister
import numpy as np
from qiskit import Aer
from qiskit import execute

def combine(x, encoder, model, measurement=None):
    """
    Inputs:
        - x (np.ndarray): Data example to encode
        - encoder (object derived from Encode class)
        - model (object derived from Model class)
        - measurement=None (object derived from Measurement class): If none,
        model.default_measurement will be used.

    Returns:
        - (qiksit.QuantumCircuit) The full circuit which encodes x using
        the encode object and combines the encoding, model, and measurement
        into a single circuit.
    """

    # Get circuits and measurement object
    encode_circuit = encoder.circuit(x)
    if model.n_qubits == None:
        #print("Setting model.n_qubits")
        model.n_qubits = encoder.n_qubits(x)
    model_circuit = model.circuit()
    if measurement == None:
        measurement = model.default_measurement()

    # Make sure circuits have the same number of qubits
    n_qubits_difference = int(abs(encode_circuit.num_qubits - model_circuit.num_qubits))
    if n_qubits_difference != 0:
        qr = QuantumRegister(n_qubits_difference)
        if encode_circuit.num_qubits < model_circuit.num_qubits:
            encode_circuit.add_register(qr)
        else:
            model_circuit.add_register(qr)

    # Combine circuits
    full_circuit = QuantumCircuit.compose(encode_circuit, model_circuit)

    # Rotate basis
    if measurement.rotate == True:
        measurement.rotate_basis(full_circuit)

    # Add classical registers and measurement operations
    measurement.add_measurements(full_circuit, measurement.qubits)

    return full_circuit

def run_data_point(x, encoder, model, measurement, backend, n_shots):
    full_circuit = combine(x, encoder, model, measurement)
    counts = execute(full_circuit, backend).result().get_counts(full_circuit)
    output = measurement.output(counts)
    prediction = measurement.output(counts)

    return prediction

def run(X, encoder, model,
        measurement=None,
        backend=Aer.get_backend("qasm_simulator"),
        n_shots=1024):
    """
    Determines size of dataset and calls run_data_point to get individual predictions.
    
    Inputs:
        - X (np.ndarray): Data set of examples
        - encoder (object derived from Encode class)
        - model (object derived from Model class)
        - measurement=None (object derived from Measurement class): If none,
        model.default_measurement will be used.
        - backend=Aer.get_backend("qasm_simulator") (qiskit backend):
        - n_shots=1024 (int): Number of times to run the circuit

    Returns:
        - predicitions (np.ndarray)
    """

    if measurement == None:
        measurement = model.default_measurement()

    if len(X.shape) == 1:
        # Single data example
        dataset_size = 1
        predictions = np.array(run_data_point(X, encoder, model, measurement, backend, n_shots))

    else:
        # Dataset
        dataset_size = X.shape[0]
        predictions = np.zeros(X.shape[0])
        for i, x in enumerate(X):
            predictions[i] = run_data_point(x, encoder, model, measurement, backend, n_shots)

    return predictions



import numpy as np

class Observable():

    def __init__(self, matrix, name="Obs", eigenvalues=None, eigenvectors=None):
        """
        Note: columns encode eigenvectors
        """

        self.matrix = matrix
        self.name = name

        if eigenvalues is None:
            set_values = True
        else:
            self.eigenvalues = eigenvalues
            set_values = False

        if eigenvectors is None:
            set_vectors = True
        else:
            self.eigenvectors = eigenvectors
            set_vectors = False

        if set_values or set_vectors:
            self.set_eigen(set_values, set_vectors)


    def set_eigen(self, set_values=False, set_vectors=False):
        """
        Performs eigen-decompositions on self.matrix.

        Sets self.eigenvalues if set_bvalues=True
        Sets self.eigenvectors if set_vectors=True
        """

        e_vals, e_vecs = np.linalg.eig(self.matrix)

        if set_values == True:
            self.eigenvalues = e_vals

        if set_vectors == True:
            #e_vecs = 1/np.sqrt(2) * np.array([[1, 1], [1, -1]])
            self.eigenvectors = e_vecs


    @staticmethod
    def X():
        X_matrix = np.array([[0, 1],
                             [1, 0]])

        e_vals = np.array([1, -1])

        e_vecs = 1/np.sqrt(2) * np.array([[1, 1],
                                          [1, -1]])

        return Observable(matrix=X_matrix, name="X", eigenvalues=e_vals, eigenvectors=e_vecs)


    @staticmethod
    def Y():
        Y_matrix = np.array([[0, -1j],
                             [1j, 0]])

        e_vals = np.array([1, -1])

        e_vecs = 1/np.sqrt(2) * np.array([[-1j, 1],
                                          [1, -1j]])

        return Observable(matrix=Y_matrix, name="Y", eigenvalues=e_vals, eigenvectors=e_vecs)


    @staticmethod
    def Z():
        Z_matrix = np.array([[1, 0],
                             [0, -1]])

        e_vals = np.array([1, -1])

        e_vecs = np.array([[1, 0],
                           [0, 1]])

        return Observable(matrix=Z_matrix, name="Z", eigenvalues=e_vals, eigenvectors=e_vecs)

    @staticmethod
    def H():
        H_matrix = 1/np.sqrt(2) * np.array([[1, 1],
                                            [1, -1]])

        e_vals = np.array([1, -1])

        e_vecs = np.array([[np.cos(np.pi/8), -np.sin(np.pi/8)],
                           [np.sin(np.pi/8), np.cos(np.pi/8)]])

        return Observable(matrix=H_matrix, name="H", eigenvalues=e_vals, eigenvectors=e_vecs)
